/* Copyright (C) 2012 Kristian Lauszus, TKJ Electronics. All rights reserved.

   This software may be distributed and modified under the terms of the GNU
   General Public License version 2 (GPL2) as published by the Free Software
   Foundation and appearing in the file GPL2.TXT included in the packaging of
   this file. Please note that GPL2 Section 2[b] requires that all works based
   on this software must also be made publicly available under the terms of
   the GPL2 ("Copyleft").

   Contact information
   -------------------

   Kristian Lauszus, TKJ Electronics
Web      :  http://www.tkjelectronics.com
e-mail   :  kristianl@tkjelectronics.com
*/

#include "BTD.h"
#include "SPP.h"

const uint32_t BTD::BTD_CONTROL_PIPE = 0;
const uint32_t BTD::BTD_EVENT_PIPE = 1;
const uint32_t BTD::BTD_DATAIN_PIPE = 2;
const uint32_t BTD::BTD_DATAOUT_PIPE = 3;

BTD::BTD(USBHost *p, const char* name, const char* pin) :
  pUsb(p), // Pointer to USB class instance - mandatory
  hci_last_invoked(0),
  bAddress(0), // Device address - mandatory
  bNumConfigurations(0), // Available configurations
  maxPktSize(0),
  bNumEP(1), // If config descriptor needs to be parsed
  qNextPollTime(0), // Reset NextPollTime
  pollInterval(BTD_ENDPOINT_POLLING_INTERVAL), // FIXME: Originally zero
  bPollEnable(false), // Don't start polling before dongle is connected
  bAuthState(true)
{/*{{{*/
  for (uint8_t i = 0; i < BTD_NUMSERVICES; i++)
    btService[i] = NULL;

  clearAllVariables(); // Set all variables, endpoint structs etc. to default values

  setName(name);
  setPin(pin);

  if (pUsb) // Register in USB subsystem
    pUsb->RegisterDeviceClass(this); // Set devConfig[] entry
}/*}}}*/

void BTD::setName(const char *name) {/*{{{*/
  memset(btdName, 0, sizeof(btdName));
  memmove(btdName, const_cast<char *>(name), min(strlen(name),sizeof(btdName)-1));
  btdName[sizeof(btdName)-1] = 0;
}/*}}}*/

void BTD::setPin(const char *pin) {/*{{{*/
  memset(btdPin, 0, sizeof(btdPin));
  memmove(btdPin, const_cast<char *>(pin), min(strlen(pin),sizeof(btdPin)-1));
  btdPin[sizeof(btdPin)-1] = 0;
}/*}}}*/

uint32_t BTD::ConfigureDevice(uint32_t parent, uint32_t port, uint32_t lowspeed) {/*{{{*/

  const uint8_t constBufSize = sizeof (USB_DEVICE_DESCRIPTOR);
  uint8_t buf[constBufSize];
  uint32_t rcode;
  UsbDevice *p = NULL;
  EpInfo *previous_ep = NULL;

  clearAllVariables(); // Set all variables, endpoint structs etc. to default values

  AddressPool &addrPool = pUsb->GetAddressPool(); // Get memory address of USB device address pool
  printf("BTD ConfigureDevice\r\n");

  if (bAddress) { // Check if address has already been assigned to an instance
    printf("Address in use\r\n");
    return USB_ERROR_CLASS_INSTANCE_ALREADY_IN_USE;
  }

  p = addrPool.GetUsbDevicePtr(0); // Get pointer to pseudo device with address 0 assigned
  if (!p) {
    printf("Address not found\r\n");
    return USB_ERROR_ADDRESS_NOT_FOUND_IN_POOL;
  }

  if (!p->epinfo) {
    printf("epinfo is null\r\n");
    return USB_ERROR_EPINFO_IS_NULL;
  }

  TRACE_USBHOST(printf("    -> Invoke getDevDescr, bufsize = %u\r\n", constBufSize);)

  // Save old pointer to EP_RECORD of address 0
  // Temporarily assign new pointer to epInfo to p->epinfo in order to avoid toggle inconsistence
  // What IS this?!
  previous_ep = p->epinfo;
  p->epinfo   = epInfo;
  p->lowspeed = lowspeed;
  rcode       = pUsb->getDevDescr(0, 0, constBufSize, (uint8_t*)buf); // Get device descriptor - addr, ep, nbytes, data
  p->epinfo   = previous_ep;

  if (rcode)
    goto FailGetDevDescr;

  TRACE_USBHOST(printf("    -> Invoke addrPoll.AllocAddress(0x%04.4x,false,0x%04.4x)\r\n",parent,port);)
  bAddress = addrPool.AllocAddress(parent, false, port); // Allocate new address according to device class

  if (!bAddress) {
    TRACE_USBHOST(printf("Out of address space\r\n");)
    return USB_ERROR_OUT_OF_ADDRESS_SPACE_IN_POOL;
  }

  // Assign Endpoint 0 Maximum Packet Size from device descriptor
  maxPktSize = (uint8_t)((USB_DEVICE_DESCRIPTOR*)buf)->bMaxPacketSize0;
  bNumConfigurations = (uint8_t)((USB_DEVICE_DESCRIPTOR*)buf)->bNumConfigurations;

  epInfo[ BTD_CONTROL_PIPE ].maxPktSize = maxPktSize; 
  epInfo[ BTD_CONTROL_PIPE ].uotghs_ptype = UOTGHS_HSTPIPCFG_PTYPE_CTRL;

  VID = ((USB_DEVICE_DESCRIPTOR*)buf)->idVendor;
  PID = ((USB_DEVICE_DESCRIPTOR*)buf)->idProduct;

  printf("BTD::ConfigureDevice       Vendor ID: 0x%4.4x\r\n", VID);
  printf("BTD::ConfigureDevice      Product ID: 0x%4.4x\r\n", PID);
  printf("BTD::ConfigureDevice        bAddress: 0x%4.4x\r\n", bAddress);
  printf("BTD::ConfigureDevice  Configurations: 0x%04.4x\r\n", bNumConfigurations);
  printf("BTD::ConfigureDevice Max Packet Size: %u\r\n", maxPktSize);

  return 0;

FailGetDevDescr:
  printf("BTD::ConfigureDevice  Unable to get dev descriptor %d\r\n", rcode);
  Release();
  return rcode;
};/*}}}*/

uint32_t BTD::Init(uint32_t parent, uint32_t port, uint32_t lowspeed) {/*{{{*/
  
  char * bt_name = (char *)&btdName;
  char * bt_pin = (char *)&btdPin; 
  uint32_t rcode;
  uint32_t alloc_status;

  TRACE_USBHOST(printf("\r\n%lu - BTD::Init()\r\n", GetTickCount());)

  pUsb->ClearAddressPool();

  if ( !bNumConfigurations ) {
    rcode = ConfigureDevice(parent, port, lowspeed);
    printf("BTD::Init - No configurations found; reinitialize result = %u, configurations = %u!\r\n", rcode, bNumConfigurations);
  }

  AddressPool &addrPool = pUsb->GetAddressPool();
  TRACE_USBHOST(printf("BTD Init, configurations %d, bAddress %llu\r\n", bNumConfigurations, bAddress);)

  UsbDevice *p = addrPool.GetUsbDevicePtr(bAddress); // Get pointer to assigned address record

  if (!p) {
    printf("Address %llu not found\r\n", bAddress);
    return USB_ERROR_ADDRESS_NOT_FOUND_IN_POOL;
  }

  delay(300); // Assign new address to the device

  rcode = pUsb->setAddr(0, 0, bAddress); // Assign new address to the device

  if (rcode) {
    printf("BTD::Init - setAddr err %x\r\n",rcode);
    p->lowspeed = false;
    goto Fail;
  }
  TRACE_USBHOST(printf("bAddress: 0x%x\r\n",bAddress);)

  p->lowspeed = false;

  p = addrPool.GetUsbDevicePtr(bAddress); // Get pointer to assigned address record

  if (!p) {
    printf("Address not found\r\n");
    return USB_ERROR_ADDRESS_NOT_FOUND_IN_POOL;
  }

  p->lowspeed = lowspeed;

  rcode = pUsb->setEpInfoEntry(bAddress, 1, epInfo); // Assign epInfo to epinfo pointer - only EP0 is known

  if (rcode)
    goto FailSetDevTblEntry;

  // Check if attached device is a Bluetooth dongle and fill endpoint data structure
  // First interface in the configuration must have Bluetooth assigned Class/Subclass/Protocol
  // And 3 endpoints - interrupt-IN, bulk-IN, bulk-OUT, not necessarily in this order
  TRACE_USBHOST(printf("Checking device VID 0x%04.4x PID 0x%04.4x. Configurations: %u; initial endpoint count %u\r\n", VID, PID, bNumConfigurations, bNumEP);)

  for (uint8_t i = 0; i < bNumConfigurations; i++) {
    // Volume 04 Host Controller Interface [Transport Layer]
    // Revision 1.2 or later
    // Part C, Ch. 2.3
    TRACE_USBHOST(printf("=> Configuration %u/%u:\r\n", i, bNumConfigurations);)
    ConfigDescParser<USB_CLASS_WIRELESS_CTRL, WI_SUBCLASS_RF, WI_PROTOCOL_BT, CP_MASK_COMPARE_ALL> confDescrParser(this);
    rcode = pUsb->getConfDescr(bAddress, 0, i, &confDescrParser);

    if (rcode) // Check error code
      goto FailGetConfDescr;

    if (bNumEP >= BTD_MAX_ENDPOINTS) // All endpoints extracted
      break;
  }

  if (bNumEP < BTD_MAX_ENDPOINTS) {
    goto FailUnknownDevice;
  }

  // Assign epInfo to epinfo pointer - this time all 3 endpoints
  rcode = pUsb->setEpInfoEntry(bAddress, bNumEP, epInfo);
  if (rcode)
    goto FailSetDevTblEntry;

  // Execute USB_REQUEST_SET_CONFIGURATION
  rcode = pUsb->setConf(bAddress, epInfo[ BTD_CONTROL_PIPE ].deviceEpNum, bConfNum);
  if (rcode)
    goto FailSetConfDescr;

  TRACE_USBHOST(printf("Checking device VID 0x%04.4x PID 0x%04.4x. Configurations: %u; Final endpoint count %u\r\n", VID, PID, bNumConfigurations, bNumEP);)

  for (uint8_t endpoint = 1 ; endpoint < bNumEP ; endpoint++ ) {
    alloc_status = UHD_Pipe_Alloc(
      endpoint,
      epInfo[endpoint].deviceEpNum & 0x7f,
      epInfo[endpoint].uotghs_ptype,
      epInfo[endpoint].deviceEpNum & 0x80 ? UOTGHS_DEVEPTCFG_EPDIR_IN : UOTGHS_DEVEPTCFG_EPDIR_OUT,
      epInfo[endpoint].maxPktSize,
      pollInterval,
      UOTGHS_HSTPIPCFG_PBK_1_BANK
    );
    if (alloc_status != endpoint) {
      TRACE_USBHOST(printf("/!\\ BTD::EndpointXtract : EP%u [0x%02.2x] allocation error: 0x%04.4x\r\n", endpoint, epInfo[endpoint].deviceEpNum, alloc_status);)
    }
    else {
      epInfo[endpoint].hostPipeNum = endpoint;
      TRACE_USBHOST(printf("/!\\ BTD::EndpointXtract : EP%u [0x%02.2x] allocation OK\r\n", endpoint, epInfo[endpoint].deviceEpNum );)
    }
  }

  printf("Bluetooth Dongle LM Configuration Done. USB Config Descriptor size: %d\r\n", sizeof(USB_CONFIGURATION_DESCRIPTOR));

  disconnect();

  hci_counter         = 0;
  hci_state           = HCI_INIT_STATE;
  connectionWait      = false;
  bPollEnable         = true;

  return 0; // Successful configuration

  /* diagnostic messages */
FailGetDevDescr:
  printf("Failed to get device descriptor. rcode = 0x%04.4x\r\n", rcode);
  goto Fail;

FailSetDevTblEntry:
  printf("Failed to set device table entry. rcode = 0x%04.4x\r\n", rcode);
  goto Fail;

FailGetConfDescr:
  printf("Failed to get device configuration descriptor. rcode = 0x%04.4x\r\n", rcode);
  goto Fail;

FailSetConfDescr:
  printf("Failed to set device configuration descriptor. rcode = 0x%04.4x\r\n", rcode);
  goto Fail;

FailUnknownDevice:
  printf("Unknown device VID 0x%04.4x PID %04.4x\r\n", VID, PID);
  pUsb->setAddr(bAddress, 0, 0); // Reset address
  rcode = USB_DEV_CONFIG_ERROR_DEVICE_NOT_SUPPORTED;
Fail:
  printf("BTD Init Failed, error code: %x\r\n", rcode);
  Release();
  return rcode;
}/*}}}*/

void BTD::clearAllVariables() {/*{{{*/
  uint8_t i;
  for (i = 0; i < BTD_NUMSERVICES; i++) {
    if (!(NULL == btService[i])) {
      btService[i]->Reset(); // Reset all Bluetooth services
    }
  }
  for (i = 0; i < BTD_MAX_ENDPOINTS; i++) {
    epInfo[i].deviceEpNum     = 0;
    epInfo[i].maxPktSize      = (i) ? 0 : 8;
    epInfo[i].epAttribs       = 0;
    epInfo[i].bmNakPower      = (i) ? USB_NAK_NOWAIT : USB_NAK_MAX_POWER;
    epInfo[i].bDescriptorType = 0;
    epInfo[i].uotghs_ptype    = 0;
  }
  pollInterval       = BTD_ENDPOINT_POLLING_INTERVAL;
  bPollEnable        = false; // Don't start polling before dongle is connected
  bAddress           = 0;     // Clear device address
  hci_last_invoked   = NULL;
  bNumConfigurations = 0;     // Available configurations
  maxPktSize         = 0;
  bNumEP             = 1;     // Must have to be reset to 1
  qNextPollTime      = 0;     // Reset next poll time
  hci_counter        = 0;
}/*}}}*/

/* Extracts interrupt-IN, bulk-IN, bulk-OUT endpoint information from config descriptor */
void BTD::EndpointXtract(uint32_t conf, uint32_t iface, uint32_t alt, uint32_t proto, const USB_ENDPOINT_DESCRIPTOR *pep) {/*{{{*/

  uint8_t index = 0;

  if (alt) {// Wrong interface - by BT spec, no alt setting
    TRACE_USBHOST(printf("> Alternate #%u, ignore.\r\n",alt);)
    // PrintEndpointDescriptor(pep);
    return;
  }

  TRACE_USBHOST(printf("> Extract endpoint information conf %u, iface %u, proto %u\r\n",conf,iface,proto);)
  bConfNum = conf;

  if ((pep->bmAttributes & 0x03) == 3 && (pep->bEndpointAddress & 0x80) == 0x80) {
    // Interrupt In endpoint found
    index = BTD_EVENT_PIPE;
    epInfo[index].bmNakPower = USB_NAK_NOWAIT;
    epInfo[index].uotghs_ptype = UOTGHS_HSTPIPCFG_PTYPE_INTRPT;
    TRACE_USBHOST(printf("> Found event pipe, index #%u\r\n",index);)
  } else {
    // Bulk endpoint found
    if ((pep->bmAttributes & 0x02) == 2)
      index = ((pep->bEndpointAddress & 0x80) == 0x80) ? BTD_DATAIN_PIPE : BTD_DATAOUT_PIPE;
    else {
      // TRACE_USBHOST(printf("!!!!!!> Skipping pipe endpoint %04.4x attrs #%04.4x\r\n", pep->bEndpointAddress, pep->bEndpointAddress);)
      // PrintEndpointDescriptor(pep);
      return;
    }
    epInfo[index].uotghs_ptype = UOTGHS_HSTPIPCFG_PTYPE_BLK;
    TRACE_USBHOST(printf("> Found bulk pipe type %u index #%u\r\n", pep->bEndpointAddress, index);)
  }

  // Fill the rest of endpoint data structure
  epInfo[index].maxPktSize  = (uint8_t)pep->wMaxPacketSize;
  // USB spec allows up to 127 devices; retain direction bit in 0x80
  epInfo[index].deviceEpNum = pep->bEndpointAddress;
  epInfo[index].bDescriptorType = pep->bDescriptorType;

  PrintEndpointDescriptor(pep);

  // Set the polling interval as the largest polling interval obtained from endpoints
  if (pollInterval < pep->bInterval)
    pollInterval = pep->bInterval;
  TRACE_USBHOST(printf("> Polling interval: Now #%u\r\n", pollInterval);)
  bNumEP++;
}/*}}}*/

void BTD::PrintEndpointDescriptor(const USB_ENDPOINT_DESCRIPTOR* ep_ptr) {/*{{{*/
  char * typeList[] = {
    "UNSPEC",
    "DEVICE",
    "CONFIGURATION",
    "STRING",
    "INTERFACE",
    "ENDPOINT",
    "DEVICE_QUALIFIER",
    "OTHER_SPEED",
    "INTERFACE_POWER",
    "OTG"
  };
  printf("Endpoint descriptor:\r\n");
  printf("\t    Length:\t0x%02.2x\r\n",ep_ptr->bLength);
  printf("\t      Type:\t0x%02.2x USB_DESCRIPTOR_%s\r\n",ep_ptr->bDescriptorType, (ep_ptr->bDescriptorType < 10 ? typeList[ep_ptr->bDescriptorType] : typeList[0]));
  printf("\t EndptAddr:\t0x%02.2x (%s)\r\n",ep_ptr->bEndpointAddress,ep_ptr->bEndpointAddress & 0x80 ? "In" : "Out");
  printf("\tAttributes:\t0x%02.2x\r\n",ep_ptr->bmAttributes);
  printf("\tMaxPktSize:\t0x%02.2x\r\n",ep_ptr->wMaxPacketSize);
  printf("\tPoll Intrv:\t0x%02.2x\r\n",ep_ptr->bInterval);
}/*}}}*/

/* Performs a cleanup after failed Init() attempt */
uint32_t BTD::Release() {/*{{{*/
  clearAllVariables(); // Set all variables, endpoint structs etc. to default values
  pUsb->GetAddressPool().FreeAddress(bAddress);
  return 0;
}/*}}}*/

uint32_t BTD::Poll() {/*{{{*/
  if (!bPollEnable) return 0;
  // FIXME: Freeze on error, no further polling allowed.
  if (pUsb->getUsbTaskState() != USB_STATE_RUNNING) return 0;
  if (qNextPollTime > GetTickCount()) return 0;
  
  // Don't poll if shorter than polling interval

  HCI_event_task(); // poll the HCI event pipe

  for (uint8_t i = 0; i < BTD_NUMSERVICES; i++)
    if (btService[i])
      btService[i]->Run(epInfo[ BTD_DATAIN_PIPE ]);

  qNextPollTime = GetTickCount() + pollInterval; // Set new poll time

  return 0;
}/*}}}*/

   
void BTD::HCI_event_task() {/*{{{*/

  // Note: Spec 4.0 Vol. 2, 5.4.4 Interrupt-IN packet limit: 255 bytes.
  uint32_t rcode = 1; // To force HCI_event_task() to bail out if we have not entered an appropriate state.
  uint32_t MAX_BUFFER_SIZE = min(epInfo[ BTD_EVENT_PIPE ].maxPktSize,HCI_BUFFERSIZE); // Request more than 16 bytes anyway, the Transfer routine will take care of this

  hcibuf[0] = 0;

  rcode = pUsb->Transfer(bAddress, epInfo[ BTD_EVENT_PIPE ].deviceEpNum, &MAX_BUFFER_SIZE, hcibuf);

#ifdef ENABLE_RETRY_ON_TIMEOUT
  if ( ( rcode == USB_ERROR_TRANSFER_TIMEOUT ) && (hci_last_invoked != NULL) ) {
    printf("/!\\ BTD::HCI_event_task : Timeout. Retrying last command\r\n");
    qNextPollTime = GetTickCount() + pollInterval;
    while ( qNextPollTime > GetTickCount() ) {}
    (this->*BTD::hci_last_invoked)();
    HCI_Task();
    return;
  }
#endif

  if (!rcode) {/*{{{*/
    // Consistent with Spec 4.0 Vol. 2, 5.4.4+
    // Check for errors
    switch (hcibuf[0]) //switch on event type
    {
      case EV_COMMAND_COMPLETE:
        if (!hcibuf[5]) { // Check if command succeeded
          hci_event_flag |= HCI_FLAG_CMD_COMPLETE; // set command complete flag
          if ((hcibuf[3] == 0x01) && (hcibuf[4] == 0x10)) { // parameters from read local version information
            hci_version = hcibuf[6]; // Used to check if it supports 2.0+EDR - see http://www.bluetooth.org/Technical/AssignedNumbers/hci.htm
            hci_event_flag |= HCI_FLAG_READ_VERSION;
          } else if ((hcibuf[3] == 0x09) && (hcibuf[4] == 0x10)) { // parameters from read local bluetooth address
            for (uint8_t i = 0; i < 6; i++)
              my_bdaddr[i] = hcibuf[6 + i];
            hci_event_flag |= HCI_FLAG_READ_BDADDR;
          }
        }
        break;

      case EV_COMMAND_STATUS:
        if (hcibuf[2]) { // show status on serial if not OK
          printf("HCI Command Failed: 0x%02.2x %02.2x %02.2x\r\n", hcibuf[2], hcibuf[4], hcibuf[6]);
        }
        break;

      case EV_INQUIRY_COMPLETE:
        if (inquiry_counter >= 5) {
          inquiry_counter = 0;
          TRACE_USBHOST(printf("Inquiry complete; enter HCI_SCANNING_STATE\r\n\r\n");)
          hci_state = HCI_SCANNING_STATE;
        }
        inquiry_counter++;
        break;

      case EV_INQUIRY_RESULT:
        if (hcibuf[2]) { // Check that there is more than zero responses
#ifdef EXTRADEBUG
          TRACE_USBHOST(printf("Number of responses: %d\r\n",hcibuf[2]);)
#endif
          for (uint8_t i = 0; i < hcibuf[2]; i++) {
            hci_event_flag |= HCI_FLAG_NONEMPTY_INQUIRY_RESULT;
#ifdef EXTRADEBUG
            TRACE_USBHOST(printf("Class of device: 0x%02.2x %02.2x %02.2x\r\n",hcibuf[6 + 8 * hcibuf[2] + 3 * i], hcibuf[5 + 8 * hcibuf[2] + 3 * i], hcibuf[4 + 8 * hcibuf[2] + 3 * i] );)
#endif
          }
        }
        break;

      case EV_CONNECT_COMPLETE:
        hci_event_flag |= HCI_FLAG_CONNECT_EVENT;
        if (!hcibuf[2]) { // check if connected OK
          hci_handle = hcibuf[3] | ((hcibuf[4] & 0x0F) << 8); // store the handle for the ACL connection
          hci_event_flag |= HCI_FLAG_CONN_COMPLETE; // set connection complete flag
        }
        else {
          printf("Connection Failed\r\n");
        }
        break;

      case EV_DISCONNECT_COMPLETE:
        if (!hcibuf[2]) { // check if disconnected OK
          hci_event_flag |= HCI_FLAG_DISCONN_COMPLETE; // set disconnect command complete flag
          hci_event_flag &= ~HCI_FLAG_CONN_COMPLETE; // clear connection complete flag
        }
        break;

      case EV_REMOTE_NAME_COMPLETE:
        if (!hcibuf[2]) { // check if reading is OK
          for (uint8_t i = 0; i < sizeof(remote_name); i++)
            remote_name[i] = hcibuf[9 + i];
          hci_event_flag |= HCI_FLAG_REMOTE_NAME_COMPLETE;
        }
        break;

      case EV_INCOMING_CONNECT:
        disc_bdaddr[0] = hcibuf[2];
        disc_bdaddr[1] = hcibuf[3];
        disc_bdaddr[2] = hcibuf[4];
        disc_bdaddr[3] = hcibuf[5];
        disc_bdaddr[4] = hcibuf[6];
        disc_bdaddr[5] = hcibuf[7];
#ifdef EXTRADEBUG
        TRACE_USBHOST(printf("Class of device: 0x%02.2x %02.2x %02.2x\r\n",hcibuf[10], hcibuf[9], hcibuf[8] );)
#endif
        hci_event_flag |= HCI_FLAG_INCOMING_REQUEST;
        break;

      case EV_PIN_CODE_REQUEST:
        if (0 < strlen(btdPin)) {
          printf("Bluetooth pin is set to: %s\r\n", &btdPin);
          hci_pin_code_request_reply();
        } else {
          printf("No pin was set\r\n");
          hci_pin_code_negative_request_reply();
        }
        break;

      case EV_LINK_KEY_REQUEST:
        printf("Received Key Request\r\n");
        hci_link_key_request_negative_reply();
        break;

      case EV_AUTHENTICATION_COMPLETE:
        printf("Pairing successful\r\n");
        break;
        /* We will just ignore the following events */
      case EV_NUM_COMPLETE_PKT:
      case EV_ROLE_CHANGED:
      case EV_PAGE_SCAN_REP_MODE:
      case EV_LOOPBACK_COMMAND:
      case EV_DATA_BUFFER_OVERFLOW:
      case EV_CHANGE_CONNECTION_LINK:
      case EV_MAX_SLOTS_CHANGE:
      case EV_QOS_SETUP_COMPLETE:
      case EV_LINK_KEY_NOTIFICATION:
      case EV_ENCRYPTION_CHANGE:
      case EV_READ_REMOTE_VERSION_INFORMATION_COMPLETE:
        break;
      default:
        if (hcibuf[0] != 0x00) {
          printf("Unmanaged HCI Event: 0x%02.2x\r\n",hcibuf[0]);
        }
        break;
    } // switch
  }/*}}}*/
  HCI_task();
}/*}}}*/

/* Poll Bluetooth and print result */
void BTD::HCI_task() {/*{{{*/
  // HCI state machine implementation 

  ticks = GetTickCount();

  switch (hci_state) {

    case HCI_INIT_STATE:
      hci_counter++;
      if (hci_counter > RESET_DEFERRAL_LOOPS) { // wait until we have looped x times to clear any old events
        hci_reset();
        // >= 2 * 10 millisecond pause after reset (USB 2.0 9.2.6.2)
        qNextPollTime = GetTickCount() + 100;
        hci_counter = 0;
      }
      break;

    case HCI_RESET_STATE:
      if (hci_cmd_complete) {
        hci_read_local_supported_commands();
      }
      break;

    case HCI_READ_LOCAL_SUPPORTED_COMMANDS:
      if (hci_cmd_complete) {
        hci_write_class_of_device();
      }
      break;

    case HCI_CLASS_STATE:
      if (hci_cmd_complete || noresp_limit) {
        hci_set_default_event_mask();
      }
      break;

    case HCI_SET_EVTMASK_STATE:
      if (hci_cmd_complete || noresp_limit) {
        hci_clear_event_filter();
      }
      break;

    case HCI_CLEAR_EVENTS_STATE:
      if (hci_cmd_complete) {
        hci_set_default_event_filter();
      }
      break;

    case HCI_SET_EVENTS_STATE:
      if (hci_cmd_complete || noresp_limit) {
        hci_write_authentication_enable(bAuthState);
      }
      break;

    case HCI_WRITE_AUTHENTICATION_ENABLE:
      if (hci_cmd_complete) {
        hci_read_bdaddr();
      }
      break;

    case HCI_BDADDR_STATE:
      if (hci_read_bdaddr_complete || noresp_limit) {
        printf("%lu - Local Bluetooth Address: ", ticks);
        for (int8_t i = 5; i > 0; i--) {
          printf("%02.2x:",my_bdaddr[i]);
        }
        printf("%02.2x\r\n",my_bdaddr[0]);
        hci_read_local_version_information();
      }
      break;

    case HCI_LOCAL_VERSION_STATE: // The local version is used by the PS3BT class
      if (hci_read_version_complete) {
        if (0 < strlen(&(btdName[0]))) {
          hci_set_local_name(&(btdName[0]));
        }
        else {
          hci_write_scan_enable();
        }
      }
      break;

    case HCI_SET_NAME_STATE:
      if (hci_cmd_complete) {
        hci_write_scan_enable();
      }
      break;

    case HCI_SCANNING_STATE:
      printf("%lu - Wait For Incoming Connection Request\r\n", ticks);
      connectionWait = true;
      hci_state = HCI_CONNECT_IN_STATE;
      break;

    case HCI_CONNECT_IN_STATE:
      if (hci_incoming_connect_request) {
        connectionWait = false;
        printf("%lu - Incoming Connection Request\r\n", ticks);
        hci_remote_name();
      }
      break;

    case HCI_REMOTE_NAME_STATE:
      if (hci_remote_name_complete) {
        printf("%lu - Remote Name: ", ticks);
        for (uint8_t i = 0; i < 30; i++) {
          if (remote_name[i] == NULL)
            break;
          printf("%c",remote_name[i]);
        }
        printf("\r\n");
        hci_accept_connection();
      }
      break;

    case HCI_CONNECTED_STATE:
      if (hci_connect_complete) {
        printf("%lu - Connected to Device: ", ticks);
        for (int8_t i = 5; i > 0; i--) {
          printf("%02.2x",disc_bdaddr[i]);
          printf(":");
        }
        printf("%02.2x\r\n",disc_bdaddr[0]);
        // Clear these flags for a new connection
        l2capConnectionClaimed = false;
        sdpConnectionClaimed = false;
        rfcommConnectionClaimed = false;

        hci_event_flag = 0;
        hci_state = HCI_DONE_STATE;
      }
      break;

    case HCI_DONE_STATE:
      //hci_counter++;
      //if (hci_counter > 1000) { // Wait until we have looped 1000 times to make sure that the L2CAP connection has been started
      //  hci_counter = 0;
      //  hci_state = HCI_SCANNING_STATE;
      //}
      break;

    case HCI_DISCONNECT_STATE:
      if (hci_disconnect_complete) {
        TRACE_USBHOST(printf("%lu - HCI Disconnected from Device\r\n", ticks);)
        hci_event_flag = 0; // Clear all flags
        // Reset all buffers
        memset(&hcibuf, 0, sizeof(hcibuf));
        hci_state = HCI_SCANNING_STATE;
      }
      break;
    default:
      break;
  }
  pUsb->setHciState(hci_state);
}/*}}}*/

/************************************************************/
/*                    HCI Commands                        */
/************************************************************/

void BTD::HCI_Command(uint8_t* data, uint16_t nbytes, uint16_t read_length) {/*{{{*/
  // Use the control pipe to send HCI-specific commands
  // (Core Spec 4.0 Vol. 4 Part B Sec. 2.2)
  uint32_t endpoint = epInfo[ BTD_CONTROL_PIPE ].deviceEpNum;
  uint32_t errorcode;
  hci_event_flag &= ~HCI_FLAG_CMD_COMPLETE;
  // TRACE_USBHOST(printf("--> HCI %u-byte command to device 0x%02.2x endpoint %llu\r\n",nbytes,bAddress,endpoint);)
  errorcode = pUsb->ctrlReq(
    bAddress,
    endpoint,
    bmREQ_HCI_OUT,  // bmReqType Request direction.
    0x00,           // bRequest Request type.
    0x00,           // wValLo Value low.
    0x00,           // wValHi Value high.
    0x00,           // wInd Index field.
    nbytes,         // total Request length.
    nbytes,         // nbytes Number of bytes to read.
    data,           // dataptr Data pointer.
    NULL            // p USB class reader.
  );
  if ( errorcode ) { 
    printf("--> HCI %u-byte command to device 0x%02.2x endpoint %llu result %llu\r\n",nbytes,bAddress,endpoint,errorcode);
  }
}/*}}}*/

void BTD::hci_read_local_supported_commands() {/*{{{*/
  printf("%lu - Local supported commands read.\r\n", ticks);
  hci_state = HCI_READ_LOCAL_SUPPORTED_COMMANDS;
  hcibuf[0] = 0x02; // HCI OCF = 0x0002, Read Local Supported Commands 
  hcibuf[1] = 0x04 << 2; // HCI OGF = 4 (Control and Baseband, Core Rev 4.0 Vol 2.E.7.4)
  hcibuf[2] = 0x00; // No parameters 

  HCI_Command(hcibuf, 3, 65);
}/*}}}*/

void BTD::hci_set_default_event_mask() {/*{{{*/
  hci_state  = HCI_SET_EVTMASK_STATE;
  hcibuf[0]  = 0x01; // HCI OCF = 0x0001, Set event mask 
  hcibuf[1]  = 0x03 << 2; // HCI OGF = 3 (Control and Baseband, Core Rev 4.0 Vol 2.E.7.3)
  hcibuf[2]  = 0x08;
  hcibuf[3]  = 0xFF; // Enable everything if all-ones
  hcibuf[4]  = 0x9F;
  hcibuf[5]  = 0xFB;
  hcibuf[6]  = 0x07;
  hcibuf[7]  = 0xF8;
  hcibuf[8]  = 0x1F;
  hcibuf[9]  = 0x2F; // 20: Pairing event  
  hcibuf[10] = 0x2F; // 20: Pairing event  
  printf("%lu - Setting event mask 0x%02.2X 0x%02.2X 0x%02.2X 0x%02.2X 0x%02.2X 0x%02.2X 0x%02.2X 0x%02.2X\r\n", ticks, hcibuf[3], hcibuf[4], hcibuf[5], hcibuf[6], hcibuf[7], hcibuf[8], hcibuf[9], hcibuf[10]);

  HCI_Command(hcibuf, 11);
}/*}}}*/

void BTD::hci_clear_event_filter() {/*{{{*/
  printf("%lu - Clearing event filters.\r\n", ticks);
  hci_set_event_filter( EV_FILTER_TYPE_CLEAR_ALL );
  hci_state = HCI_CLEAR_EVENTS_STATE;
}/*}}}*/

void BTD::hci_set_default_event_filter() {/*{{{*/
  printf("%lu - Setting default event filters.\r\n", ticks);
  uint8_t accept_all[1];
  accept_all[0] = EV_FILTER_COND_CONN_ALL_WITH_ROLESWITCH;
  hci_set_event_filter(
    EV_FILTER_TYPE_CONNECTION_SETUP,
    EV_FILTER_COND_CONN_ACCEPT_ALL,
    1,
    &(accept_all[0])
  );
  hci_state = HCI_SET_EVENTS_STATE;
}/*}}}*/

void BTD::hci_set_event_filter(uint8_t filter_type, uint8_t filter_cond, uint8_t param_len, uint8_t * filter_params ) {/*{{{*/
  // Set event filter, see Bluetooth Specification V 4.0, Vol. 2, Part E, Sec. 7.3.3.
  hci_state = HCI_SET_EVENTS_STATE;
  hcibuf[0] = 0x05; // HCI OCF = 0x0005
  hcibuf[1] = 0x03 << 2; // HCI OGF = 3 (Control and Baseband, Core Rev 4.0 Vol 2.E.7.3)
  hcibuf[2] = 1;
  hcibuf[3] = filter_type;
  uint16_t len = 0;
  switch ( filter_type ) {
    case EV_FILTER_TYPE_CLEAR_ALL: 
      len = 4;
      break;
    case EV_FILTER_TYPE_INQUIRY_RESULT:
      hcibuf[4] = filter_cond;
      switch ( filter_cond ) {/*{{{*/
        case EV_FILTER_COND_INQ_ACCEPT_ALL:
          len = 5;
          break;
        case EV_FILTER_COND_INQ_CLASS_RESPONDER:
        case EV_FILTER_COND_INQ_ADDR_RESPONDER:
        default: return;
      }/*}}}*/
      break;
    case EV_FILTER_TYPE_CONNECTION_SETUP:
      hcibuf[4] = filter_cond;
      switch ( filter_cond ) {/*{{{*/
        case EV_FILTER_COND_CONN_ACCEPT_ALL:
          if ( param_len == 1 ) {
            hcibuf[5] = filter_params[0]; // AUTO_ACCEPT flag
            len = 6;
          } else return;
          break;
        case EV_FILTER_COND_CONN_CLASS_MATCHED:
          if ( param_len == 7 ) {
            len = 5;
            while ( param_len ) {
              len++;
              hcibuf[len + param_len - 1] = filter_params[param_len-1];
              param_len--;
            }
          } else return;
        case EV_FILTER_COND_CONN_ADDR_MATCHED:
        default: return;
      }/*}}}*/
      break;
    default: return;
  }
  HCI_Command(hcibuf, len);
}/*}}}*/

void BTD::hci_reset() {/*{{{*/
  printf("%lu - Executing reset.\r\n", ticks);
  hci_event_flag = 0; // Clear all the flags
  hci_state = HCI_RESET_STATE;
  hcibuf[0] = 0x03; // HCI_Reset Vol 4.2.E.7.3.2 
  hcibuf[1] = 0x03 << 2; // HCI OGF = 3 (Control and Baseband, Core Rev 4.0 Vol 2.E.7.3)
  hcibuf[2] = 0x00; // parameter length = 0

  HCI_Command(hcibuf, 3, 1);
  hci_last_invoked = &BTD::hci_reset;
}/*}}}*/

void BTD::hci_write_scan_enable() {/*{{{*/
  hci_event_flag &= ~HCI_FLAG_INCOMING_REQUEST;
  hci_state = HCI_SCANNING_STATE;
  hcibuf[0] = 0x1A; // HCI OCF = 1A
  hcibuf[1] = 0x03 << 2; // HCI OGF = 3
  hcibuf[2] = 0x01; // parameter length = 1
  if (0 < strlen(btdName))
    hcibuf[3] = 0x03; // Inquiry Scan enabled. Page Scan enabled.
  else
    hcibuf[3] = 0x02; // Inquiry Scan disabled. Page Scan enabled.

  HCI_Command(hcibuf, 4, 1);
  hci_last_invoked = &BTD::hci_write_scan_enable;
}/*}}}*/

void BTD::hci_write_scan_disable() {/*{{{*/
  hcibuf[0] = 0x1A; // HCI OCF = 1A
  hcibuf[1] = 0x03 << 2; // HCI OGF = 3
  hcibuf[2] = 0x01; // parameter length = 1
  hcibuf[3] = 0x00; // Inquiry Scan disabled. Page Scan disabled.

  HCI_Command(hcibuf, 4);
  hci_last_invoked = &BTD::hci_write_scan_disable;
}/*}}}*/

void BTD::hci_read_bdaddr() {/*{{{*/
  hci_state = HCI_BDADDR_STATE;
  hcibuf[0] = 0x09; // HCI OCF = 9
  hcibuf[1] = 0x04 << 2; // HCI OGF = 4
  hcibuf[2] = 0x00;

  HCI_Command(hcibuf, 3, 7);
  hci_last_invoked = &BTD::hci_read_bdaddr;
}/*}}}*/

void BTD::hci_read_local_version_information() {/*{{{*/
  hci_state = HCI_LOCAL_VERSION_STATE;
  hcibuf[0] = 0x01; // HCI OCF = 1
  hcibuf[1] = 0x04 << 2; // HCI OGF = 4
  hcibuf[2] = 0x00;

  HCI_Command(hcibuf, 3, 9);
  hci_last_invoked = &BTD::hci_read_local_version_information;
}/*}}}*/

void BTD::hci_accept_connection() {/*{{{*/
  hci_event_flag &= ~HCI_FLAG_CONN_COMPLETE;
  hci_state = HCI_CONNECTED_STATE;
  hcibuf[0] = 0x09; // HCI OCF = 9
  hcibuf[1] = 0x01 << 2; // HCI OGF = 1
  hcibuf[2] = 0x07; // parameter length 7
  hcibuf[3] = disc_bdaddr[0]; // 6 octet bdaddr
  hcibuf[4] = disc_bdaddr[1];
  hcibuf[5] = disc_bdaddr[2];
  hcibuf[6] = disc_bdaddr[3];
  hcibuf[7] = disc_bdaddr[4];
  hcibuf[8] = disc_bdaddr[5];
  hcibuf[9] = 0x00; //switch role to master

  HCI_Command(hcibuf, 10);
  hci_last_invoked = &BTD::hci_accept_connection;
}/*}}}*/

void BTD::hci_remote_name() {/*{{{*/
  hci_event_flag &= ~HCI_FLAG_REMOTE_NAME_COMPLETE;
  hci_state = HCI_REMOTE_NAME_STATE;
  hcibuf[0] = 0x19; // HCI OCF = 19
  hcibuf[1] = 0x01 << 2; // HCI OGF = 1
  hcibuf[2] = 0x0A; // parameter length = 10
  hcibuf[3] = disc_bdaddr[0]; // 6 octet bdaddr
  hcibuf[4] = disc_bdaddr[1];
  hcibuf[5] = disc_bdaddr[2];
  hcibuf[6] = disc_bdaddr[3];
  hcibuf[7] = disc_bdaddr[4];
  hcibuf[8] = disc_bdaddr[5];
  hcibuf[9] = 0x01; //Page Scan Repetition Mode
  hcibuf[10] = 0x00; //Reserved
  hcibuf[11] = 0x00; //Clock offset - low byte
  hcibuf[12] = 0x00; //Clock offset - high byte

  HCI_Command(hcibuf, 13);
  hci_last_invoked = &BTD::hci_remote_name;
}/*}}}*/

void BTD::hci_set_local_name(const char* name) {/*{{{*/
  printf("%lu - Assigning name: '%s'\r\n", ticks, name);
  hci_state = HCI_SET_NAME_STATE;
  hcibuf[0] = 0x13; // HCI OCF = 13
  hcibuf[1] = 0x03 << 2; // HCI OGF = 3
  hcibuf[2] = strlen(name) + 1; // parameter length = the length of the string + end byte
  uint8_t i;
  for (i = 0; i < strlen(name); i++)
    hcibuf[i + 3] = name[i];
  hcibuf[i + 3] = 0x00; // End of string

  HCI_Command(hcibuf, 4 + strlen(name));
  hci_last_invoked = NULL; 
}/*}}}*/

void BTD::hci_inquiry() {/*{{{*/
  hci_state = HCI_INQUIRY_STATE; 
  hci_event_flag &= ~HCI_FLAG_NONEMPTY_INQUIRY_RESULT;
  hcibuf[0] = 0x01;
  hcibuf[1] = 0x01 << 2; // HCI OGF = 1
  hcibuf[2] = 0x05; // Parameter Total Length = 5
  hcibuf[3] = 0x33; // LAP: General/Unlimited Inquiry Access Code (GIAC = 0x9E8B33) - see https://www.bluetooth.org/Technical/AssignedNumbers/baseband.htm
  hcibuf[4] = 0x8B;
  hcibuf[5] = 0x9E;
  hcibuf[6] = 0x30; // Inquiry time N * 1.28s; 48 => 61.44 sec (maximum)
  hcibuf[7] = 0x0A; // 0x0A: 10 number of responses; 0x00: Unlimited responses

  HCI_Command(hcibuf, 8);
  hci_last_invoked = &BTD::hci_inquiry;
}/*}}}*/

void BTD::hci_inquiry_cancel() {/*{{{*/
  hcibuf[0] = 0x02;
  hcibuf[1] = 0x01 << 2; // HCI OGF = 1
  hcibuf[2] = 0x00; // Parameter Total Length = 0

  HCI_Command(hcibuf, 3);
  hci_last_invoked = &BTD::hci_inquiry_cancel;
}/*}}}*/

void BTD::hci_connect() {/*{{{*/
  hci_event_flag &= ~(HCI_FLAG_CONN_COMPLETE | HCI_FLAG_CONNECT_EVENT);
  hcibuf[0] = 0x05;
  hcibuf[1] = 0x01 << 2; // HCI OGF = 1
  hcibuf[2] = 0x0D; // parameter Total Length = 13
  hcibuf[3] = disc_bdaddr[0]; // 6 octet bdaddr
  hcibuf[4] = disc_bdaddr[1];
  hcibuf[5] = disc_bdaddr[2];
  hcibuf[6] = disc_bdaddr[3];
  hcibuf[7] = disc_bdaddr[4];
  hcibuf[8] = disc_bdaddr[5];
  hcibuf[9] = 0x18; // DM1 or DH1 may be used
  hcibuf[10] = 0xCC; // DM3, DH3, DM5, DH5 may be used
  hcibuf[11] = 0x01; // Page repetition mode R1
  hcibuf[12] = 0x00; // Reserved
  hcibuf[13] = 0x00; // Clock offset
  hcibuf[14] = 0x00; // Invalid clock offset
  hcibuf[15] = 0x00; // Do not allow role switch

  HCI_Command(hcibuf, 16);
  hci_last_invoked = &BTD::hci_connect;
}/*}}}*/

void BTD::hci_pin_code_request_reply() {/*{{{*/

  if (0 == strlen(&(btdPin[0]))) {
    TRACE_USBHOST(printf("No PIN configured; sending negative reply\r\n");)
    hci_pin_code_negative_request_reply();
    return;
  } 
  hcibuf[0] = 0x0D; // HCI OCF = 0D
  hcibuf[1] = 0x01 << 2; // HCI OGF = 1
  hcibuf[2] = 0x17; // parameter length 23
  hcibuf[3] = disc_bdaddr[0]; // 6 octet bdaddr
  hcibuf[4] = disc_bdaddr[1];
  hcibuf[5] = disc_bdaddr[2];
  hcibuf[6] = disc_bdaddr[3];
  hcibuf[7] = disc_bdaddr[4];
  hcibuf[8] = disc_bdaddr[5];

  hcibuf[9] = strlen(const_cast<char *>(&btdPin[0])); // Length of pin
  uint8_t i;
  for (i = 0; i < hcibuf[9]; i++) // The maximum size of the pin is 16
    hcibuf[i + 10] = btdPin[i];
  for (; i < 16; i++)
    hcibuf[i + 10] = 0x00; // The rest should be 0

  HCI_Command(hcibuf, 26);
  hci_last_invoked = &BTD::hci_pin_code_request_reply;
}/*}}}*/

void BTD::hci_pin_code_negative_request_reply() {/*{{{*/
  hcibuf[0] = 0x0E; // HCI OCF = 0E
  hcibuf[1] = 0x01 << 2; // HCI OGF = 1
  hcibuf[2] = 0x06; // parameter length 6
  hcibuf[3] = disc_bdaddr[0]; // 6 octet bdaddr
  hcibuf[4] = disc_bdaddr[1];
  hcibuf[5] = disc_bdaddr[2];
  hcibuf[6] = disc_bdaddr[3];
  hcibuf[7] = disc_bdaddr[4];
  hcibuf[8] = disc_bdaddr[5];

  HCI_Command(hcibuf, 9);
  hci_last_invoked = &BTD::hci_pin_code_negative_request_reply;
}/*}}}*/

void BTD::hci_link_key_request_negative_reply() {/*{{{*/
  hcibuf[0] = 0x0C; // HCI OCF = 0C
  hcibuf[1] = 0x01 << 2; // HCI OGF = 1
  hcibuf[2] = 0x06; // parameter length 6
  hcibuf[3] = disc_bdaddr[0]; // 6 octet bdaddr
  hcibuf[4] = disc_bdaddr[1];
  hcibuf[5] = disc_bdaddr[2];
  hcibuf[6] = disc_bdaddr[3];
  hcibuf[7] = disc_bdaddr[4];
  hcibuf[8] = disc_bdaddr[5];

  HCI_Command(hcibuf, 9);
  hci_last_invoked = &BTD::hci_link_key_request_negative_reply;
}/*}}}*/

void BTD::hci_write_authentication_enable(bool v) {/*{{{*/
  printf("%lu - Set Require Authentication Enable flag %s.\r\n", ticks, v ? "true" : "false");
  hci_state = HCI_WRITE_AUTHENTICATION_ENABLE;
  hcibuf[0] = 0x20; // HCI OCF = 20 
  hcibuf[1] = 0x03 << 2; // HCI OGF = 1
  hcibuf[2] = 0x01; // parameter length = 1
  hcibuf[3] = v ? 0x01 : 0x00; 

  HCI_Command(hcibuf, 4);
  hci_last_invoked = NULL;
}/*}}}*/

void BTD::hci_authentication_request() {/*{{{*/
  printf("%lu - Accept authentication request, hci_handle 0x%04.4X.\r\n", ticks, hci_handle);
  hcibuf[0] = 0x11; // HCI OCF = 11
  hcibuf[1] = 0x01 << 2; // HCI OGF = 1
  hcibuf[2] = 0x02; // parameter length = 2
  hcibuf[3] = (uint8_t)(hci_handle & 0xFF); //connection handle - low byte
  hcibuf[4] = (uint8_t)((hci_handle >> 8) & 0x0F); //connection handle - high byte

  HCI_Command(hcibuf, 5);
  hci_last_invoked = &BTD::hci_authentication_request;
}/*}}}*/

void BTD::hci_disconnect(uint16_t handle) {/*{{{*/
  // This is called by the different services
  hci_event_flag &= ~HCI_FLAG_DISCONN_COMPLETE;
  hcibuf[0] = 0x06; // HCI OCF = 6
  hcibuf[1] = 0x01 << 2; // HCI OGF = 1
  hcibuf[2] = 0x03; // parameter length = 3
  hcibuf[3] = (uint8_t)(handle & 0xFF); //connection handle - low byte
  hcibuf[4] = (uint8_t)((handle >> 8) & 0x0F); //connection handle - high byte
  hcibuf[5] = 0x13; // reason

  HCI_Command(hcibuf, 6);
  hci_last_invoked = NULL;
}/*}}}*/

void BTD::hci_write_class_of_device() {/*{{{*/
  // See http://bluetooth-pentest.narod.ru/software/bluetooth_class_of_device-service_generator.html
  printf("%lu - Assigning device class.\r\n", ticks);
  hci_state = HCI_CLASS_STATE;
  hcibuf[0] = 0x24; // HCI OCF = 0x24 
  hcibuf[1] = 0x03 << 2; // HCI OGF = 3
  hcibuf[2] = 0x03; // parameter length = 3
  hcibuf[3] = 0x14; // Palm-sized PC/PDA 
  hcibuf[4] = 0x01; // Computer
  hcibuf[5] = 0xC2; // Networking, telephony, information

  HCI_Command(hcibuf, 6);
  hci_last_invoked = &BTD::hci_write_class_of_device;
}/*}}}*/

/*******************************************************************
 *                                                                 *
 *                        HCI ACL Data Packet                      *
 *                                                                 *
 *   buf[0]          buf[1]          buf[2]          buf[3]
 *   0       4       8    11 12      16              24            31 MSB
 *  .-+-+-+-+-+-+-+-|-+-+-+-|-+-|-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-.
 *  |      HCI Handle       |PB |BC |       Data Total Length       |   HCI ACL Data Packet
 *  .-+-+-+-+-+-+-+-|-+-+-+-|-+-|-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-.
 *
 *   buf[4]          buf[5]          buf[6]          buf[7]
 *   0               8               16                            31 MSB
 *  .-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-.
 *  |            Length             |          Channel ID           |   Basic L2CAP header
 *  .-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-.
 *
 *   buf[8]          buf[9]          buf[10]         buf[11]
 *   0               8               16                            31 MSB
 *  .-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-.
 *  |     Code      |  Identifier   |            Length             |   Control frame (C-frame)
 *  .-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-|-+-+-+-+-+-+-+-.   (signaling packet format)
 */

/************************************************************/
/*                    L2CAP Commands                        */
/************************************************************/
void BTD::L2CAP_Command(uint16_t handle, uint8_t* data, uint8_t nbytes, uint8_t channelLow, uint8_t channelHigh) {/*{{{*/
  uint8_t buf[8 + nbytes];
  buf[0] = (uint8_t)(handle & 0xff); // HCI handle with PB,BC flag
  buf[1] = (uint8_t)(((handle >> 8) & 0x0f) | 0x20);
  buf[2] = (uint8_t)((4 + nbytes) & 0xff); // HCI ACL total data length
  buf[3] = (uint8_t)((4 + nbytes) >> 8);
  buf[4] = (uint8_t)(nbytes & 0xff); // L2CAP header: Length
  buf[5] = (uint8_t)(nbytes >> 8);
  buf[6] = channelLow;
  buf[7] = channelHigh;

  for (uint16_t i = 0; i < nbytes; i++) // L2CAP C-frame
    buf[8 + i] = data[i];
  uint32_t transfer_length = (8 + nbytes);
  uint32_t rcode = pUsb->Transfer(bAddress, epInfo[ BTD_DATAOUT_PIPE ].deviceEpNum, &transfer_length, buf);
  if (rcode) {
    delay(100); // This small delay prevents it from overflowing if it fails
#ifdef DEBUG_USB_HOST
    TRACE_USBHOST(printf("Error sending L2CAP message: 0x%08.8x\r\n", rcode);)
    TRACE_USBHOST(printf(" - Channel ID: 0x%02.2x %02.2x",channelHigh,channelLow);)
#endif
  }
}/*}}}*/

void BTD::l2cap_connection_request(uint16_t handle, uint8_t rxid, uint8_t* scid, uint16_t psm) {/*{{{*/
  l2capoutbuf[0] = L2CAP_CMD_CONNECTION_REQUEST; // Code
  l2capoutbuf[1] = rxid; // Identifier
  l2capoutbuf[2] = 0x04; // Length
  l2capoutbuf[3] = 0x00;
  l2capoutbuf[4] = (uint8_t)(psm & 0xff); // PSM
  l2capoutbuf[5] = (uint8_t)(psm >> 8);
  l2capoutbuf[6] = scid[0]; // Source CID
  l2capoutbuf[7] = scid[1];

  L2CAP_Command(handle, l2capoutbuf, 8);
}/*}}}*/

void BTD::l2cap_connection_response(uint16_t handle, uint8_t rxid, uint8_t* dcid, uint8_t* scid, uint8_t result) {/*{{{*/
  l2capoutbuf[0] = L2CAP_CMD_CONNECTION_RESPONSE; // Code
  l2capoutbuf[1] = rxid; // Identifier
  l2capoutbuf[2] = 0x08; // Length
  l2capoutbuf[3] = 0x00;
  l2capoutbuf[4] = dcid[0]; // Destination CID
  l2capoutbuf[5] = dcid[1];
  l2capoutbuf[6] = scid[0]; // Source CID
  l2capoutbuf[7] = scid[1];
  l2capoutbuf[8] = result; // Result: Pending or Success
  l2capoutbuf[9] = 0x00;
  l2capoutbuf[10] = 0x00; // No further information
  l2capoutbuf[11] = 0x00;

  L2CAP_Command(handle, l2capoutbuf, 12);
}/*}}}*/

void BTD::l2cap_config_request(uint16_t handle, uint8_t rxid, uint8_t* dcid) {/*{{{*/
  l2capoutbuf[0] = L2CAP_CMD_CONFIG_REQUEST; // Code
  l2capoutbuf[1] = rxid; // Identifier
  l2capoutbuf[2] = 0x08; // Length
  l2capoutbuf[3] = 0x00;
  l2capoutbuf[4] = dcid[0]; // Destination CID
  l2capoutbuf[5] = dcid[1];
  l2capoutbuf[6] = 0x00; // Flags
  l2capoutbuf[7] = 0x00;
  l2capoutbuf[8] = 0x01; // Config Opt: type = MTU (Maximum Transmission Unit) - Hint
  l2capoutbuf[9] = 0x02; // Config Opt: length
  l2capoutbuf[10] = 0xFF; // MTU
  l2capoutbuf[11] = 0xFF;

  L2CAP_Command(handle, l2capoutbuf, 12);
}/*}}}*/

void BTD::l2cap_config_response(uint16_t handle, uint8_t rxid, uint8_t* scid) {/*{{{*/
  l2capoutbuf[0] = L2CAP_CMD_CONFIG_RESPONSE; // Code
  l2capoutbuf[1] = rxid; // Identifier
  l2capoutbuf[2] = 0x0A; // Length
  l2capoutbuf[3] = 0x00;
  l2capoutbuf[4] = scid[0]; // Source CID
  l2capoutbuf[5] = scid[1];
  l2capoutbuf[6] = 0x00; // Flag
  l2capoutbuf[7] = 0x00;
  l2capoutbuf[8] = 0x00; // Result
  l2capoutbuf[9] = 0x00;
  l2capoutbuf[10] = 0x01; // Config
  l2capoutbuf[11] = 0x02;
  l2capoutbuf[12] = 0xA0;
  l2capoutbuf[13] = 0x02;

  L2CAP_Command(handle, l2capoutbuf, 14);
}/*}}}*/

void BTD::l2cap_disconnection_request(uint16_t handle, uint8_t rxid, uint8_t* dcid, uint8_t* scid) {/*{{{*/
  l2capoutbuf[0] = L2CAP_CMD_DISCONNECT_REQUEST; // Code
  l2capoutbuf[1] = rxid; // Identifier
  l2capoutbuf[2] = 0x04; // Length
  l2capoutbuf[3] = 0x00;
  l2capoutbuf[4] = dcid[0];
  l2capoutbuf[5] = dcid[1];
  l2capoutbuf[6] = scid[0];
  l2capoutbuf[7] = scid[1];

  L2CAP_Command(handle, l2capoutbuf, 8);
}/*}}}*/

void BTD::l2cap_disconnection_response(uint16_t handle, uint8_t rxid, uint8_t* dcid, uint8_t* scid) {/*{{{*/
  l2capoutbuf[0] = L2CAP_CMD_DISCONNECT_RESPONSE; // Code
  l2capoutbuf[1] = rxid; // Identifier
  l2capoutbuf[2] = 0x04; // Length
  l2capoutbuf[3] = 0x00;
  l2capoutbuf[4] = dcid[0];
  l2capoutbuf[5] = dcid[1];
  l2capoutbuf[6] = scid[0];
  l2capoutbuf[7] = scid[1];

  L2CAP_Command(handle, l2capoutbuf, 8);
}/*}}}*/

void BTD::l2cap_information_response(uint16_t handle, uint8_t rxid, uint8_t infoTypeLow, uint8_t infoTypeHigh) {/*{{{*/
  l2capoutbuf[0] = L2CAP_CMD_INFORMATION_RESPONSE; // Code
  l2capoutbuf[1] = rxid; // Identifier
  l2capoutbuf[2] = 0x08; // Length
  l2capoutbuf[3] = 0x00;
  l2capoutbuf[4] = infoTypeLow;
  l2capoutbuf[5] = infoTypeHigh;
  l2capoutbuf[6] = 0x00; // Result = success
  l2capoutbuf[7] = 0x00; // Result = success
  l2capoutbuf[8] = 0x00;
  l2capoutbuf[9] = 0x00;
  l2capoutbuf[10] = 0x00;
  l2capoutbuf[11] = 0x00;

  L2CAP_Command(handle, l2capoutbuf, 12);
}/*}}}*/

/* PS3 Commands - only set Bluetooth address is implemented in this library */
void BTD::setBdaddr(uint8_t* bdaddr) {/*{{{*/
  /* Set the internal Bluetooth address */
  uint8_t buf[8];
  buf[0] = 0x01;
  buf[1] = 0x00;

  for (uint8_t i = 0; i < 6; i++)
    buf[i + 2] = bdaddr[5 - i]; // Copy into buffer, has to be written reversed

  // bmRequest = Host to device (0x00) | Class (0x20) | Interface (0x01) = 0x21, bRequest = Set Report (0x09), Report ID (0xF5), Report Type (Feature 0x03), interface (0x00), datalength, datalength, data)
  pUsb->ctrlReq(bAddress, epInfo[BTD_CONTROL_PIPE].deviceEpNum, bmREQ_HID_OUT, HID_REQUEST_SET_REPORT, 0xF5, 0x03, 0x00, 8, 8, buf, NULL);
}/*}}}*/

void BTD::setMoveBdaddr(uint8_t* bdaddr) {/*{{{*/
  /* Set the internal Bluetooth address */
  uint8_t buf[11];
  buf[0] = 0x05;
  buf[7] = 0x10;
  buf[8] = 0x01;
  buf[9] = 0x02;
  buf[10] = 0x12;

  for (uint8_t i = 0; i < 6; i++)
    buf[i + 1] = bdaddr[i];

  // bmRequest = Host to device (0x00) | Class (0x20) | Interface (0x01) = 0x21, bRequest = Set Report (0x09), Report ID (0x05), Report Type (Feature 0x03), interface (0x00), datalength, datalength, data)
  pUsb->ctrlReq(bAddress, epInfo[BTD_CONTROL_PIPE].deviceEpNum, bmREQ_HID_OUT, HID_REQUEST_SET_REPORT, 0x05, 0x03, 0x00, 11, 11, buf, NULL);
}/*}}}*/


